package com.revfad.search;

import java.util.*;

/**
 * I represent a queue of documents waiting to be used.
 * I am thread-safe.
 * <p>
 * This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike License. To view a copy of this license, visit <a href="http://creativecommons.org/licenses/by-nc-sa/2.0/">http://creativecommons.org/licenses/by-nc-sa/2.0/</a> or send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.
 *
 * @author David Faden, dfaden@gmail.com
 */
public class DocumentQueue {

private Vector queue;

private boolean killed;

public DocumentQueue() {
    queue = new Vector();
    killed = false;
}

public DocumentQueue(Vector documents) {
    queue = (Vector) documents.clone();
    killed = false;
}

/**
 * The next element in the queue. If there is no such element,
 * this method blocks until one becomes available. This method
 * returns null if the queue is killed.
 *
 * @return The next element in the queue or null if the queue
 * has been killed.
 */
public Document next() {
    synchronized (queue) {
        while (queue.isEmpty() && !killed) {
            try {
                queue.wait();
            }
            catch (InterruptedException ie) {
                return null;
            }
        }
        
        if (killed) {
            return null;
        }
        
        Document result = (Document) queue.firstElement();
        queue.removeElementAt(0);
        return result;
    }
}

/**
 * @param newLast The document to add to the end of the queue.
 */
public void addLast(Document newLast) {
    synchronized (queue) {
        queue.addElement(newLast);
        queue.notify();
    }
}


public boolean isEmpty() {
    //Given that Vectors are synchronized, do we need to
    //put the following in a synchronized block?

    synchronized (queue) {
        return queue.isEmpty();
    }
}

/**
 * Unblock any callers waiting on next(), returning null
 * to each. The result of subsequent calls to next() is
 * undefined.
 */
public void kill() {
    synchronized (queue) {
        killed = true;
        queue.notifyAll();
    }
}

}