package com.revfad.search;

/**
 * I hold information about a search match.
 * <p>
 * This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike License. To view a copy of this license, visit <a href="http://creativecommons.org/licenses/by-nc-sa/2.0/">http://creativecommons.org/licenses/by-nc-sa/2.0/</a> or send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.
 */
public final class Match {

private final String pattern;
private final Document document;
private final String context;
private final String nearestAnchor;

/**
 * @param pattern The pattern that completed the match, not necessarily including all of the components of the match.
 * @param document The document containing the match.
 * @param context The context of the match. May be null.
 * @param nearestAnchor The closest anchor preceding the match in the document. May be null.
 */
public Match(String pattern, Document document, String context, String nearestAnchor) {
    this.pattern = pattern;
    this.document = document;
    this.context = context;
    this.nearestAnchor = nearestAnchor;
}

/**
 * The pattern that completed the match, not necessarily including all of the components of the match.
 */
public String pattern() {
    return pattern;
}

/**
 * The document containing the match.
 */
public Document document() {
    return document;
}

/**
 * The context of the match. This is text from
 * the document up to and including the match and text following
 * the match. May be null.
 */
public String context() {
    return context;
}

/**
 * The closest anchor preceding the match in the document. May be null.
 */
public String nearestAnchor() {
    return nearestAnchor;
}

private String nullOrQuoted(String s) {
    if (s == null) {
        return "null";
    }
    else {
        return "\"" + s + "\"";
    }
}

public String toString() {
    return document + " pattern = \"" + pattern() + "\"" +
        "; context = " + nullOrQuoted(context())
        + "; nearestAnchor = " + nullOrQuoted(nearestAnchor());
}

/**
 * Compare the titles and names (in that order) of
 * of the documents of this match and the given
 * match (in that order).
 *
 * @param o The Match with which to compare this one.
 * @return A number less than 0 if this match
 * is less than <tt>o</tt>, 0 if the matches are equal,
 * and a number greater than 0 if this match is
 * greater than <tt>o</tt>.
 */
public int compareTo(Object o) {
    Match match = (Match) o;
    String myTitle = document.titleOrFileNameIfNull();
    String myFileName = document.fileName();
    
    String title = match.document().titleOrFileNameIfNull();
    String fileName = match.document().fileName();
    
    int difference = myTitle.compareTo(title);
    if (difference == 0) {
        difference = myFileName.compareTo(fileName);
    }
    return difference;
}

/**
 * Compare the last modified dates of this match and the given
 * match, using their titles to break ties.
 *
 * @param o The Match with which to compare this one.
 * @return A number less than 0 if this match's
 * document was modified more recently than <tt>o</tt>,
 * the result of comparing their titles if their dates
 * are equal, and a number greater than zero if
 * this match's document was modified less recently.
 */
public int compareDateTo(Object o) {
    Match match = (Match) o;
    long difference =
        match.document().lastModified() - document().lastModified();
    if (difference < 0) {
        return -1;
    }
    else if (difference > 0) {
        return 1;
    }
    else {
        return compareTo(o);
    }
}

private boolean bothNullOrEqual(Object o1, Object o2) {
    if (o1 == null) {
        return o2 == null;
    }
    else {
        return o1.equals(o2);
    }
}

/**
 * Two matches are considered equal if the have equal
 * patterns, documents, and nearestAnchors. Context info
 * is ignored.
 */
public boolean equals(Object o) {
    if (!(o instanceof Match)) {
        return false;
    }
    
    Match m = (Match) o;
    return document().equals(m.document())
        && pattern().equals(m.pattern())
        //&& bothNullOrEqual(context(), m.context())
        && bothNullOrEqual(nearestAnchor(), m.nearestAnchor());
}

private int hashCodeOrOneIfNull(Object o) {
    if (o == null) {
        return 1;
    }
    else {
        return o.hashCode();
    }
}

public int hashCode() {
    return 37 * 37 * 37 * document().url().hashCode()
            + 37 * 37 * pattern().hashCode()
            + 37 * hashCodeOrOneIfNull(nearestAnchor());
}

}