package com.revfad.search;

//A partial modification record:
//
// August 16, 2000: Corrected some inaccurate comments.
//
// August 17, 2000: Removed method isSpace(char). Added the 
// isWordChar(int) method to be used as negative replacement.
// isWordChar uses table look-up rather than a loop so it should be
// substantially faster. Added method setWordChar(char, boolean) 
// to match isWordChar(int).
// I consider this to be an interim version of isWordChar - the final
// one will use some sort of hash and correctly handle the full range of
// Unicode characters.
//
// August 17, 2000: changed the names of the methods numberOfMatches() and patternLength()
// to getNumOfMatches() and getKeyLength() respectively. Documented the undocumented.
//
// December 28, 2000: added some more documentation, shifted some conditionals around
// to better reflect the expected occurrence of conditions.
//
// March 11, 2001: Changed many of the variable names to be more intuitive.
// Added code to keep track of the "position" or "index" of the SearchSieve
// in relation to the stream of char fed to it. Changed the constructor to take
// a String as its argument rather than a char array. Strings are safer.
// Strings are more convenient to handle for the caller. Using a String better
// fits the Java way of thinking.
//
// March 13, 2001: Stripped out code that allowed any two white space
// char to match -- a tab would be considered to match a carriage return.
// I guestimated that the cost of this feature outweighed the benefits. 
// Also, this behavior will not be so easy (or possible) to implement if 
// and when SearchSieves are switched to use the Knuth-Morris-Pratt 
// algorithm internally. See addChar().
//
// June 25, 2004. Stripped out all of the code not essential to carrying
// a particular search. SearchSieves no longer keep track of the total
// number of matches, the number of char they've been fed, or the
// position of the last match within the text stream. None of this information
// is used in the current implementation of SearchToHTML so it is simply 
// a waste to collect it.
// 
// Also, I replaced SearchSieve's homebrewed method of distinguishing between word
// and non-word char with a call to Character.isLetter. I had used the homebrewed
// method for efficiency's sake, but now internationalization is an even bigger
// concern. And the other changes to the code have probably more than made up
// for any differences in speed.
//
// June 30, 2004 Added back a hasMatched method and accompanying field.


import java.io.*;
import java.util.*;

/**
 * I sift for a pattern in the sequence of characters fed to me.
 * <p>
 * This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike License. To view a copy of this license, visit <a href="http://creativecommons.org/licenses/by-nc-sa/2.0/">http://creativecommons.org/licenses/by-nc-sa/2.0/</a> or send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.
 *
 * @author David Faden
 * @version SearchSieve 3.0 June 26, 2004
 */
public final class SearchSieve {
    
    /**
     * True if this SearchSieve has found a match since
     * its last reset.
     */
    private boolean hasMatched;
    
    /**
     * Current position in the pattern.
     */
    private int patternIndex;
    
    /**
     * String representation of the search pattern.
     */
    private String patternString;
    
    /**
     * The pattern to search for.
     */
    private char[] pattern;
    
    /**
     * The number of characters in the pattern.
     */
    private int patternLength;
    
    /**
     * Determines whether a match can be part of a larger word.
     */
    private boolean exactMatchesOnly;
  
    /**
     * The previous character to be examined.
     */
    private char prevc;
  
    /**
     * Store characters for backtracking.
     */
    private char[] buffer;
    
    /**
     * The number of characters in the buffer.
     */
    private int bufferLength;
    
    /**
     * Determines whether this SearchSieve is buffering to backtrack.
     */
    private boolean buffering;
    
    /**
     * Storage for swaps between char arrays.
     */
    private char[] swapSpace;
  
   /**
    * Create a new SearchSieve with the given pattern, not searching
    * for exact matches only.
    *
    * @param patternString - The pattern to search for.
    */
    public SearchSieve(String patternString) {
        this(patternString, false);
    }
  
    /**
     * Create a new SearchSieve that represents the same pattern as
     * original.
     */
    public SearchSieve(SearchSieve original) {
        this(original.patternString, original.exactMatchesOnly);
    }
  
    /**
     * Create a new SearchSieve.
     *
     * @param patternString - The pattern to search for.
     * @param exactMatchesOnly - If true, matches must be surrounded by non-word characters.
     */
    public SearchSieve(String patternString, boolean exactMatchesOnly) {
        if (patternString == null) {
            throw new IllegalArgumentException("null patternString");
        }
        if (patternString.length() == 0) {
            throw new IllegalArgumentException("zero length patternString");
        }
        setPattern(patternString, exactMatchesOnly);
    }
  
    /**
     * Set the pattern this SearchSieve should search for.
     *
     * @param pattern String containing the pattern to search for.
     * @param exactMatchesOnly True if only matches surrounded by non-word
     *    characters should be accepted.
     */
    private final void setPattern(String patternString, boolean exactMatchesOnly) {
        this.exactMatchesOnly = exactMatchesOnly;
        this.patternString = patternString;
        pattern = patternString.toCharArray();
        patternLength = pattern.length;
        buffer = new char[patternLength];
        swapSpace = new char[patternLength]; 
        //swapSpace is created here to avoid the overhead of recreating it 
        //each time it is needed.
    
        reset();
    }
    
    /**
     * Reset this SearchSieve so that it acts as if addChar() has not
     * been called. The count of matches
     * is set back to zero.
     */
    public final void reset() {
        patternIndex = 0;
        prevc = '\0'; //null according to the Unicode table, nul according to others
        bufferLength = 0; //clear the buffer
        buffering = false;
        hasMatched = false;
    }
    
    
    //0009;<control>;Cc;0;S;;;;;N;HORIZONTAL TABULATION;;;;
    //000A;<control>;Cc;0;B;;;;;N;LINE FEED;;;;
    //000B;<control>;Cc;0;S;;;;;N;VERTICAL TABULATION;;;;
    //000C;<control>;Cc;0;WS;;;;;N;FORM FEED;;;;
    //000D;<control>;Cc;0;B;;;;;N;CARRIAGE RETURN;;;;
    //
    //0020;SPACE;Zs;0;WS;;;;;N;;;;;
    //00A0;NO-BREAK SPACE;Zs;0;CS;<noBreak> 0020;;;;N;NON-BREAKING SPACE;;;;
    //1680;OGHAM SPACE MARK;Zs;0;WS;;;;;N;;;;;
    //2000;EN QUAD;Zs;0;WS;2002;;;;N;;;;;
    //2001;EM QUAD;Zs;0;WS;2003;;;;N;;;;;
    //2002;EN SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
    //2003;EM SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
    //2004;THREE-PER-EM SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
    //2005;FOUR-PER-EM SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
    //2006;SIX-PER-EM SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
    //2007;FIGURE SPACE;Zs;0;WS;<noBreak> 0020;;;;N;;;;;
    //2008;PUNCTUATION SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
    //2009;THIN SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
    //200A;HAIR SPACE;Zs;0;WS;<compat> 0020;;;;N;;;;;
    //200B;ZERO WIDTH SPACE;Zs;0;BN;;;;;N;;;;;
    //202F;NARROW NO-BREAK SPACE;Zs;0;WS;<noBreak> 0020;;;;N;;;;;
    //3000;IDEOGRAPHIC SPACE;Zs;0;WS;<wide> 0020;;;;N;;;;;
    //
    //2029;PARAGRAPH SEPARATOR;Zp;0;B;;;;;N;;;;;
    //
    //2028;LINE SEPARATOR;Zl;0;WS;;;;;N;;;;;

    
    //Note that we exclude the non-breaking space char. Should we exclude it?
    private static final String HIGH_SPACE_CHARS =
            "\u1680\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009"
            + "\u200A\u200B\u202F\u3000\u2029\u2028";
    
    /**
     * Parse the query into a sequence of SearchSieves. Patterns are separated
     * by whitespace -- a space, tab, carriage return, or line feed -- unless surrounded by quotes.
     * A segment surrounded by unescaped double quotes is taken as a segment, excluding the quotes.
     * Quotes may be included in a pattern by escaping them with a backslash, as in Java: \\".
     *
     * @param query String to parse.
     * @param exactMatchesOnly Should the returned SearchSieve accept only exact matches?
     * @return The search terms parsed.
     */
    public static final SearchSieve[] parseQuery(String query, boolean exactMatchesOnly) {

        Vector sieves = new Vector();
        StreamTokenizer tokenizer = new StreamTokenizer(new StringReader(query));

        tokenizer.resetSyntax();
        
        tokenizer.wordChars(0x21, Character.MAX_VALUE);
        tokenizer.whitespaceChars(0, 0x20);
        
        final char QUOTE = '\"';
        tokenizer.quoteChar(QUOTE);
        
        try {
            while (tokenizer.nextToken() != StreamTokenizer.TT_EOF) {
                if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
                    StringTokenizer splitter = new StringTokenizer(tokenizer.sval, HIGH_SPACE_CHARS);
                    while (splitter.hasMoreTokens()) {
                        sieves.addElement(new SearchSieve(splitter.nextToken(), exactMatchesOnly));
                    }
                }
                else if (tokenizer.ttype == QUOTE) {
                    sieves.addElement(new SearchSieve(tokenizer.sval, exactMatchesOnly));   
                }
                else {
                    throw new IllegalStateException("Unrecognized token type: " + tokenizer.ttype);
                }
            }
        }
        catch (IOException ioe) {
            throw new IllegalArgumentException(ioe.toString());
        }
        SearchSieve[] result = new SearchSieve[sieves.size()];
        sieves.copyInto(result);
        return result;
    }

    
    /**
     * true if a match must be surrounded by non-word char to match.
     *
     * @return true if a match must be surrounded by non-word char to match.
     */
    public boolean getExactMatchesOnly() {
        return exactMatchesOnly;
    }
      
    /**
     * Return the String representation of the pattern
     * being searched for, or the pattern for which this
     * SearchSieve is searching, if you will.
     *
     * @return The String representation of the search pattern.
     */
    public final String getPattern() {
        return patternString;
    }
    
    /**
     * Return the number of characters in pattern (pattern) being
     * searched for.
     *
     * @return The number of characters in pattern (pattern) being
     * searched for.
     */
    public final int getPatternLength() {
        //XXX! Is useful to have this method? I think it will be
        //in case SearchSieves are later used as part of a wildcard
        //matching class.
        return patternLength;
    }

    
    //XXX! As a result of doing the conversion to lowercase in the SearchSieve,
    //we are making many, many unnecessary calls to Character.toLowerCase.
    //Maybe we should require the client to take care of the conversion
    //and just provide plain "scan" methods.
    
    /**
     * Feed all of the char of aString to this SearchSieve. Return the leftmost index in aString
     * at which a match was completed or -1 if no match was found.
     *
     * @param aString The string from which to read char.
     * @return The the leftmost index in aString at which a match was completed or -1 if
     * no match was found.
     */
    public final int scanAllLowerCase(String aString) {
        char[] text = aString.toCharArray();
        return scanAllLowerCase(text, 0, text.length);
    }

    /**
     * Feed all the char in buffer from index offset
     * to index offset + maxLength - 1 to this SearchSieve. Return the first
     * index
     */
    public final int scanAllLowerCase(char[] buffer, int offset, int maxLength) {
        int last = offset + maxLength - 1;
        int matchIndex = -1;
        int index = offset - 1;
        int remaining = maxLength;
        do {
            index = scanLowerCase(buffer, index + 1, remaining);
            if (index != -1 && matchIndex == -1) {
                matchIndex = index;
            }
            remaining = last - index;
        } while (index != -1 && index != last);
        return matchIndex;
    }

    /**
     * Feed the char of aString to this SearchSieve up to the first match, all
     * of the char if there is no match.
     * Return the leftmost index in aString at which a match was found, -1
     * if no match was found. Note that this index will be at the rightmost end
     * of the block of text constituting the match.
     *
     * @return The leftmost index in aString at which a match was found, -1
     * if no match was found.
     */
    public final int scanLowerCase(String aString) {
        char[] text = aString.toCharArray();
        return scanLowerCase(text, 0, text.length);
    }
        
    /**
     * Feed up to maxLength char of buffer to this
     * SearchSieve until a match is found or all maxLLengthif no match is found. 
     * If a match is found, return its index. Note that this index will be at the rightmost end
     * of the block of text constituting the match. If maxLength char have
     * been taken from buffer with no match, return -1.
     *
     * @return The leftmost index in buffer offset or more char
     * from index 0 which for which this SearchSieve found a match, -1
     * if there is no such index.
     */
    public final int scanLowerCase(char[] buffer, int offset, int maxLength) {
        int end = offset + maxLength;
        for (int i = offset; i < end; i++) {
            if (addChar(Character.toLowerCase(buffer[i]))) {
                return i;
            }
        }
        return -1;
    }
    
    /**
     * Returns true if the sequence of chars added to the SearchSieve,
     * including <code>c</code>, constitutes a match. For exactMatchesOnly searches,
     * the last char of a match must be a non-word char.
     * <br>
     * If the pattern is of zero length, this method always returns false.
     *
     * @param c The next char in the text stream.
     * @return True if <code>c</code> completes a match -- false, otherwise.
     */
    public final boolean addChar(char c) {
        boolean matchFound = false; //Left unchanged if nothing's found.
        boolean considerEatingBuffer = false;
        
        if (exactMatchesOnly) {
            if (patternIndex == 0) {
                if (c == pattern[0] && !Character.isLetter(prevc)) 
                    patternIndex++;
            }
            else if (patternIndex != patternLength) {
                if (c != pattern[patternIndex]) { //found no match
                    patternIndex = 0;
                    considerEatingBuffer = true;
                }
                else {
                    patternIndex++;
                }
            }
            else { //all the characters of the pattern matched
                if (!Character.isLetter(c)) { //found an "exactMatchesOnly" match
                    matchFound = true;
                }
                patternIndex = 0;
                considerEatingBuffer = true;
            }
        }
        //Inexact match check
        else {
            if (c != pattern[patternIndex]) { 
                    //No match
                    patternIndex = 0;
                    considerEatingBuffer = true;
            }
            else //Input char matches pattern
                patternIndex++;
            
            if (patternIndex == patternLength) { //All the characters matched!
                patternIndex = 0;
                considerEatingBuffer = true;
                matchFound = true;
            }
        }
        
        //Note: patternIndex of 1 indicates that one char match has been found
        if (patternIndex > 1 && !buffering && c == pattern[0]) {
            if (!exactMatchesOnly || exactMatchesOnly && !Character.isLetter(prevc)) {
                buffering = true;
                buffer[0] = c;
                bufferLength = 1;
            }
        }
        else if (buffering) {
            if (c != pattern[bufferLength]) {
                bufferLength = 0;
                buffering = false;
            }
            else {
                buffer[bufferLength] = c;
                bufferLength++;
            }
        }
        
        if (considerEatingBuffer && bufferLength > 0) 
            eatTheBuffer(); //Feed the buffer back into the
            //search system on failures and finds
            
        prevc = c;
        
        hasMatched = hasMatched || matchFound;
        
        return matchFound;
    }
  
    public boolean hasMatched() {
        return hasMatched;
    }
  
  	public String toString() {
  		return getPattern() + " " + getExactMatchesOnly();
  	}
  
    /**
     * Two SearchSieves are considered to be equal if they both represent the same
     * pattern and have the same exact matches only value. The SearchSieves' positions
     * within their patterns are not taken into account.
     */
    public boolean equals(Object o) {
        if (!(o instanceof SearchSieve)) {
            return false;
        }
        
        SearchSieve sieve = (SearchSieve)o;
        return getPattern().equals(sieve.getPattern())
                && getExactMatchesOnly() == sieve.getExactMatchesOnly();
    }
  
    public int hashCode() {
        return getPattern().hashCode() + (getExactMatchesOnly() ? 1 : 0);
    }
  
    /**
     * Feed the buffer through addChar(char) as if it were being externally entered.
     */
    private final void eatTheBuffer() {
        //For historical interest, I have preserved the original comments below:
        //Feed the buffer through addChar(char) as if it were being externally entered.
        //No match can ever be found in the buffer alone, so there is
        //no need to check the return of addChar(char).
        //Huge problem with this method which will cause fatal error...
        //Buffer itself may contain a second ambiguous char...
        //However, if the buffer doesn't match up, it is overwritten
        //When the next call to eatTheBuffer is made
        //Idea, what if swapSpace is made internal to eatTheBuffer?
        //This could lead to a major memory leak, though...
        //Fixed...breath easy now!
        //XXX! If this step were replaced with a look up table,
        //SearchSieves would roughly implement the Knuth-Morris-Pratt
        //algorithm...
        
        //Reset the search
        patternIndex = 0;
        
        System.arraycopy(buffer, 0, swapSpace, 0, bufferLength);
        int tempLength = bufferLength; //Save the value of bufferLength
        
        //Reset the buffer
        bufferLength = 0; 
        buffering = false;
    
        //Don't bother checking addChar(char)'s return value -- 
        //it can never be true in this case.
        for (int i = 0; i < tempLength; i++) 
            addChar(swapSpace[i]);
    }
  
}
