package com.revfad.search;

import java.io.*;
import java.util.*;

/**
 * I pull documents from a queue and search them, passing on the results
 * to my SearchThreadListener. I will stop searching either if
 * someone calls my stopSearching method or if the next method of
 * my queue returns null.
 * <p>
 * This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike License. To view a copy of this license, visit <a href="http://creativecommons.org/licenses/by-nc-sa/2.0/">http://creativecommons.org/licenses/by-nc-sa/2.0/</a> or send a letter to Creative Commons, 559 Nathan Abbott Way, Stanford, California 94305, USA.
 *
 * @author David Faden, dfaden@iastate.edu
 */
public final class SearchThread extends Thread implements DocumentScanner {

/**
 * I represent a search option understood by a SearchThread. Use the
 * static instances I provide as keys for the Hashtables passed to your
 * SearchThreads.
 */
public static final class Option {

/**
 * Boolean determining whether only terms in the text surrounded by non-word characters
 * may match. Defaults to Boolean.FALSE.
 */
public static final Option EXACT_MATCHES_ONLY = new Option("EXACT_MATCHES_ONLY");

/**
 * Integer specifying the number of char following a match to collect for the context.
 * Defaults to the equivalent of new Integer(15).
 */
public static final Option LEADING_CONTEXT_LENGTH = new Option("LEADING_CONTEXT_LENGTH");

/**
 * Integer specifying the number of char following a match to collect for the context.
 * Defaults to the equivalent of new Integer(0).
 */
public static final Option TRAILING_CONTEXT_LENGTH = new Option("TRAILING_CONTEXT_LENGTH");

/**
 * SearchSieve representing a pattern that must be found before we will
 * start looking for a match. If unspecified, no start pattern is used.
 */
//public static final Option START_PATTERN = new Option("START_PATTERN");

/**
 * SearchSieve representing a pattern that once matched will stop the search
 * of a document. If unspecified, no stop pattern is used.
 */
//public static final Option STOP_PATTERN = new Option("STOP_PATTERN");

/**
 * Boolean determining whether the text between matching &lt; and &gt; is ignored.
 * Defaults to Boolean.FALSE.
 */
public static final Option IGNORE_TAG_CONTENT = new Option("IGNORE_TAG_CONTENT");

/**
 * Boolean determining whether every search term must be found for the current
 * section of the document to match. Defaults to Boolean.FALSE.
 */
public static final Option MUST_MATCH_ALL = new Option("MUST_MATCH_ALL");

/**
 * Boolean determining whether I will look for multiples matches within a single
 * document. Defaults to Boolean.FALSE. 
 */
public static final Option ALLOW_MULTIPLE_MATCHES = new Option("ALLOW_MULTIPLE_MATCHES");

/**
 * A String representing the charset for all documents. If unspecified, the Documents will try to
 * guess their own character encodings. The String is assumed to be an encoding
 * name recognized by Java. As of June 26, 2004, you can find a subset of the supported
 * names at http://java.sun.com/j2se/1.3/docs/guide/intl/encoding.doc.html .
 */
//public static final Option CHARSET = new Option("CHARSET"); 

private Option(String name) {
    this.name = name;
}

public boolean getFromWithDefault(Hashtable options, boolean defaultValue) {
    Object value = options.get(this);
    if (value == null) {
        return defaultValue;
    }
    else {
        return Boolean.TRUE == value;
    }
}

public int getFromWithDefault(Hashtable options, int defaultValue) {
    Object value = options.get(this);
    if (value == null) {
        return defaultValue;
    }
    else {
        return ((Integer)value).intValue();
    }
}

public String toString() {
    return name;
}

private String name;

}

//TODO: Consider whether to add support for start patterns and stop patterns.
//Does anybody actually use them? They greatly complicate the code.

public SearchThread(SearchSieve[] query, DocumentQueue queue, SearchThreadListener listener, Hashtable options) {
    this.queue = queue;
    this.listener = listener;
    exactMatchesOnly = Option.EXACT_MATCHES_ONLY.getFromWithDefault(options, false);
    leadingContextLength = Option.LEADING_CONTEXT_LENGTH.getFromWithDefault(options, 15);
    trailingContextLength = Option.TRAILING_CONTEXT_LENGTH.getFromWithDefault(options, 0);
    ignoreTagContent = Option.IGNORE_TAG_CONTENT.getFromWithDefault(options, false);
    mustMatchAll = Option.MUST_MATCH_ALL.getFromWithDefault(options, false);
    allowMultipleMatches = Option.ALLOW_MULTIPLE_MATCHES.getFromWithDefault(options, false);
    
    numberOfPatterns = query.length;
    patterns = new SearchSieve[numberOfPatterns];
    for (int i = 0; i < query.length; i++) {
        patterns[i] = new SearchSieve(query[i]);
    }
    
    if (exactMatchesOnly) {
        contextBuffer = new CircularStringBuffer(leadingContextLength + 1);
    }
    else {
        contextBuffer = new CircularStringBuffer(leadingContextLength);
    }
    
    continueSearching = true;
}

public void run() {
    Document next = null;
    while (continueSearching && (next = queue.next()) != null) {
        listener.searching(this, next);
        try {
            inTag = false;
            contextBuffer.reset();
            anchorFinder.reset();
            nearestAnchor = null;
            anchorBuffer = null;
            searchingDocument = true;
            for (int i = 0; i < numberOfPatterns; i++) {
                patterns[i].reset();
            }
            next.beginReadingFor(this);
        }
        catch (IOException ioe) {
            //XXX! Should we add a method to the listener
            //to let it know what happened?
            ioe.printStackTrace();
        }
        if (continueSearching) {
            listener.finishedSearching(this, next);
        }
        for (int i = 0; i < numberOfPatterns; i++) {
            patterns[i].reset();
        }
    }
}

public char[] buffer(Document document) {
    //XXX! Should probably only allocate one buffer for all documents.
    int length = Math.max(4096, document.contentLength());
    length = Math.min(length, 100000);
    return new char[length];
}

public boolean shouldContinue(Document document) {
    return continueSearching && searchingDocument;
}

private boolean inTag;
private CircularStringBuffer contextBuffer;
private StringBuffer anchorBuffer;
private SearchSieve anchorFinder = new SearchSieve("<a name=\"");

public void scan(char[] buffer, int length, Document document) {
    for (int index = 0; index < length && searchingDocument; index++) {
        char c = buffer[index];
        char lowerc = Character.toLowerCase(c);
        
        if (anchorBuffer == null) {
            if (anchorFinder.addChar(lowerc)) {
                anchorBuffer = new StringBuffer();
            }
        }
        else {
           if (lowerc == '\"') {
                if (anchorBuffer.length() != 0) {
                    nearestAnchor = anchorBuffer.toString();
                    anchorBuffer = null;
                }
            }
            else {
                anchorBuffer.append(c);
            }
        }

        if (lowerc == '<') {
            inTag = true;
        }
   
        //Begin search block
        if (!ignoreTagContent || !inTag) {
            contextBuffer.append(c);
            boolean allMatched = true;
            boolean anyMatched = false;
            SearchSieve matchingSieve = null;
            for (int i = 0; i < numberOfPatterns; i++) {
                SearchSieve sieve = patterns[i];
                boolean matched = sieve.addChar(lowerc);
                if (matched) {
                    matchingSieve = sieve;
                }
                else {
                    matched = sieve.hasMatched();
                }
                allMatched = allMatched && matched;
                anyMatched = anyMatched || matched;
            }
            boolean foundMatch = (anyMatched && !mustMatchAll) || (allMatched && mustMatchAll);
            if (foundMatch) {
                String trailingContext = "";
                if (index != length - 1) {
                    StringBuffer trailingBuffer = new StringBuffer();
                    
                    boolean trailingInTag = false;
                    for (int j = index + 1;
                            j < length && trailingBuffer.length() < trailingContextLength;
                            j++) {
                        char next = buffer[j];
                        if (next == '<') {
                            trailingInTag = true;
                        }
                        
                        if (!ignoreTagContent || !trailingInTag) {
                            trailingBuffer.append(next);
                        }
                        
                        if (next == '>') {
                            trailingInTag = false;
                        }
                    }
                    
                    trailingContext = trailingBuffer.toString();
                }
                listener.foundMatch(this,
                                    new Match(matchingSieve.getPattern(), document,
                                        contextBuffer.toString() + trailingContext, nearestAnchor));
                if (allowMultipleMatches) {
                    for (int i = 0; i < numberOfPatterns; i++) {
                        patterns[i].reset();
                    }
                }
                else {
                    searchingDocument = false;
                }
            }
        }
        //End search block
        
        if (lowerc == '>') {
            inTag = false;
        }

    }


}



/**
 * Read an anchor from buffer, storing it in the nearestAnchor instance field.
 * Return the rightmost index in buffer occupied by the anchor. If buffer
 * doesn't contain the completion of any anchors, return -1.
 *
 * @param buffer The char[] to read from.
 * @param offset The index in buffer to beging reading at.
 * @param length the maximum number of char to read from buffer.
 *
 * @return The rightmost index in buffer occupied by the anchor. If buffer
 * doesn't contain the completion of any anchors, return -1.
 */
/*
private int readAnchor(char[] buffer, int offset, int length) {

    if (retrievingAnchor) {
        int max = offset + length;
        int quoteIndex = -1;
        for (int i = offset + 1; i < max; i++) {
            if (buffer[i] == '\"') {
                nearestAnchor = new String(buffer, index + 1, i - index - 1);
                retrievingAnchor = false;
                return i;
            }
        }
        //Note: we append length - 1 rather than length because
        //we're dropping the closing double quote.
        anchorBuffer.append(buffer, offset, length - 1);
        return -1;
    }

    int index = anchorFinder.scanLowerCase(buffer, offset, length);
    
    if (index == -1) {
        return -1;
    }
    
    int max = offset + length;
    retrievingAnchor = true;
    int quoteIndex = -1;
    for (int i = index + 1; i < max; i++) {
        if (buffer[i] == '\"') {
            nearestAnchor = new String(buffer, index + 1, i - index - 1);
            return i;
        }
    }

    anchorBuffer = new StringBuffer();
    anchorBuffer.append(buffer, index + 1, length - index - 1);
    retrievingAnchor = true;
}
*/


public void stopSearching() {
    continueSearching = false;
}

private DocumentQueue queue;

private SearchThreadListener listener;
 
 
//Options

private boolean exactMatchesOnly;

private int leadingContextLength;

private int trailingContextLength;

private boolean ignoreTagContent;

private boolean mustMatchAll;

private boolean allowMultipleMatches;

/**
 * The patterns to search for.
 */
private SearchSieve[] patterns;

private int numberOfPatterns;

private String nearestAnchor;

/**
 * Is this thread currently searching the current document?
 */
private boolean searchingDocument;

/**
 * Should this thread continue searching at all?
 */
private volatile boolean continueSearching;

}