package com.revfad.search.tests;

import com.revfad.search.*;

import java.io.*;
import java.net.*;
import junit.framework.*;

/**
 * If you have JUnit on your classpath, you can run this test
 * as follows:
 * java junit.swingui.TestRunner com.revfad.search.tests.DocumentTest
 */
public class DocumentTest extends TestCase {

private static class NullDocumentScanner implements DocumentScanner {

public char[] buffer(Document document) {
    return new char[4096];
}

public boolean shouldContinue(Document document) {
    return true;
}

public void scan(char[] buffer, int length, Document document) {
}

}

private Document iso_8859_1Document;
private Document macintoshDocument;
private Document gb2312Document;
private Document big5Document;
private Document cp1256Document;
private Document utf8ArabicDocument;
private Document windows1251Document;

public static Test suite() {
    return new TestSuite(DocumentTest.class);
}

public void setUp() {
    try {
        iso_8859_1Document = readLocalDocument("iso_8859_1.html");
        macintoshDocument = readLocalDocument("macintosh.html");
        gb2312Document = readLocalDocument("gb2312.html");
        big5Document = readLocalDocument("big5.html");
        cp1256Document = readLocalDocument("cp1256.html");
        utf8ArabicDocument = readLocalDocument("utf-8_arabic.html");
        windows1251Document = readLocalDocument("windows-1251.html");
    }
    catch (Exception e) {
        throw new RuntimeException(e.toString());
    }
}

private static final String BASE_ADDRESS = "http://homepages.isunet.net/faden/search_tests/";

private Document readWebDocument(String fileName) throws MalformedURLException, IOException, SecurityException {
    return readDocument(new URL(BASE_ADDRESS + fileName));
}

private Document readLocalDocument(String fileName) throws MalformedURLException, IOException, SecurityException {
    return readDocument(LocalFileNameToUrl.toUrl(fileName));
}

private Document readDocument(URL url) throws IOException, SecurityException {
    Document result = new Document(url, null);
    result.beginReadingFor(new NullDocumentScanner());
    return result;
}

public void testBuildUrlRelative() throws MalformedURLException
{
    URL base = new URL("http://www.nowhere.org/abc/def/ghi/");
    assertEquals(new URL("http://www.nowhere.org/abc/test/"),
                Document.buildUrl(base, "../../test/"));
}

public void testBuildUrlRoot() throws MalformedURLException
{
    URL base = new URL("http://www.nowhere.org/abc/def/index.html");
    assertEquals(new URL("http://www.nowhere.org/test.html"),
                Document.buildUrl(base, "/test.html"));
}

public void testBuildUrlDotSlash() throws MalformedURLException
{
    URL base = new URL("http://www.nowhere.org/abc/def/index.html");
    assertEquals(new URL("http://www.nowhere.org/abc/def/test.html"),
                Document.buildUrl(base, "./test.html"));
}

/**
 * @param docUrl String of URL to document.
 * @param baseUrl String or URL to file on same host as the document.
 */
private void checkRelativePathTo(String docUrl, String baseUrl) throws MalformedURLException {
    Document doc = new Document(new URL(docUrl), null);
    URL base = new URL(baseUrl);
    assertEquals(doc.url(), new URL(base, doc.pathRelativeTo(base)));
}

public void testRelativePathToDir() throws MalformedURLException {
    checkRelativePathTo("http://h.geocities.org/t/t/t/t/t/hope.html",
        "http://h.geocities.org/t/t/t/t/t");
}

public void testRelativePathTo() throws MalformedURLException {
    checkRelativePathTo("http://www.x.com/a/b/c/d.html",
            "http://www.x.com/a/b/g/h/p/i/k/searchresults.html");
}

public void testRelativePathToRoot() throws MalformedURLException {
    checkRelativePathTo("http://www.x.com/a/b/c/d.html",
        "http://www.x.com/oh_no/b/g/h/p/i/k/searchresults.html");
}

public void testRelativePathToWithRoot() throws MalformedURLException {
    checkRelativePathTo("http://www.x.com/a/b/c/d.html",
        "http://www.x.com/searchresults.html");
}

private void checkTitle(String expected, Document document) {
    String title = document.title();
    assertEquals(expected, title);
}

public void testGb2312Title() {
    checkTitle("\u6b47\u540e\u8bed\u4e13\u96c6\uff08\u0041\uff09", gb2312Document);
}

public void testBig5Title() {
    checkTitle("\u6b32 \u9664 \u7169 \u60f1 \u9808 \u7121 \u6211 \ufe50 \u6b77 \u7d93 \u8271 \u96e3 \u505a \u597d \u4eba \ufe54 ",
                big5Document);
}

public void testCp1256Title() {
    checkTitle(" \u0645\u0646\u0638\u0645\u0629 \u0627\u0644\u0639\u0641\u0648 \u0627\u0644\u062f\u0648\u0644\u064a\u0629 \u062a\u0642\u0631\u064a\u0631 \u0627\u0644\u0639\u0627\u0645 \u0031\u0039\u0039\u0039", cp1256Document);
}

public void testUtf8ArabicTitle() {
    checkTitle("\u0645\u062f\u062e\u0644 \u0645\u0648\u0642\u0639 \u0627\u0644\u0648\u0632\u0627\u0631\u0629\u002d \u0639\u0631\u0628\u064a", utf8ArabicDocument);
}

public void testTitle() {
    checkTitle("StrangeTalk3.5 :: Index", iso_8859_1Document);
    checkTitle("Brad's Web Page", macintoshDocument);
    checkTitle("RUSSIAN NEW ZEALAND", windows1251Document);
}

public void testFileName() {
    assertEquals("iso_8859_1.html", iso_8859_1Document.fileName());
    assertEquals("macintosh.html", macintoshDocument.fileName());
    assertEquals("gb2312.html", gb2312Document.fileName()); 
}

public void testEncoding() {
    assertEquals("ISO8859_1", iso_8859_1Document.contentEncoding());
    assertEquals("MacRoman", macintoshDocument.contentEncoding());
    assertEquals("EUC_CN", gb2312Document.contentEncoding());
    assertEquals("Big5", big5Document.contentEncoding());
    assertEquals("Cp1256", cp1256Document.contentEncoding());
    assertEquals("UTF8", utf8ArabicDocument.contentEncoding());
    assertEquals("Cp1251", windows1251Document.contentEncoding());
}

}