package com.revfad.search.tests;

import com.revfad.search.*;

import junit.framework.*;

/**
 * If you have JUnit on your classpath, you can run this test
 * as follows:
 * java junit.swingui.TestRunner com.revfad.search.tests.SearchSieveTest
 */
public class SearchSieveTest extends TestCase {

public static Test suite() {
    return new TestSuite(SearchSieveTest.class);
}

private void checkParseQuery(String query, String[] expectedPatterns, boolean exactMatchesOnly) {
    SearchSieve[] expected = new SearchSieve[expectedPatterns.length];
    for (int i = 0; i < expected.length; i++) {
        expected[i] = new SearchSieve(expectedPatterns[i], exactMatchesOnly);
    }
    SearchSieve[] actual = SearchSieve.parseQuery(query, exactMatchesOnly);
    
    assertEquals(expected.length, actual.length);
    for (int i = 0; i < expected.length; i++) {
        assertEquals(expected[i], actual[i]);
    }
}

public void testParseQuery1() {
    String[] expectedPatterns = new String[]{"Wherefore", "art", "thou", "Romeo \"Romeo\" Romeo"};
    checkParseQuery("Wherefore art thou \"Romeo \\\"Romeo\\\" Romeo\"", expectedPatterns, true);
}

public void testParseQuery2() {
    String[] expectedPatterns = new String[]{"abc", "123 123"};
    checkParseQuery("abc\"123 123\"", expectedPatterns, false);
}

public void testParseQueryChinese1() {
    String[] expectedPatterns =
        //"ni3 hao3", "wo3 kan4 shu1"
        new String[]{"\u4f60\u597d", "\u6211\u770b\u4e66"};
        
    checkParseQuery("\u4f60\u597d \u6211\u770b\u4e66", expectedPatterns, true);
}

public void testParseQueryChinese2() {
    String[] expectedPatterns =
        //"shuo1", "tuo1 le ku4 zi fang4 pi4"
        new String[]{"\u8bf4", "\u8131 \u88e4\u5b50 \u653e\u5c41"};
    checkParseQuery("\u8bf4\r\r \"\u8131 \u88e4\u5b50 \u653e\u5c41\"",
                    expectedPatterns, false);
}

public void testParseQueryIdeographicSpace() {
    String[] expectedPatterns =
        new String[]{"hi!", "123abc"};
    checkParseQuery("hi!\u3000123abc", expectedPatterns, true);
}

public void testParseQueryEmbeddedIdeographicSpace() {
    String[] expectedPatterns =
        //"zuo3", "you4", "zuo4 you4"
        new String[] {"\u5de6", "\u53f3", "\u5de6\u3000\u53f3"};
    checkParseQuery("\u5de6\u3000 \u3000\u53f3\"\u5de6\u3000\u53f3\"", expectedPatterns, false);
}

public void testParseQueryCppComment() {
    String[] expectedPatterns = new String[]{"a", "//", "b"};
    checkParseQuery("a // b", expectedPatterns, true);
}

public void testExactAtBeginning() {
    String text = "aBc\radsf9sad09f 2 s";
    SearchSieve sieve = new SearchSieve("abc", true);
    assertEquals(3, sieve.scanLowerCase(text));
}

public void testExactInMiddle1() {
    String text = "Four score and seven years...";
    SearchSieve sieve = new SearchSieve("score", true);
    assertEquals(10, sieve.scanLowerCase(text));
}

public void testExactInMiddle2() {
    String text = "Foddle pee pi! Doo dah dum!";
    SearchSieve sieve = new SearchSieve("pi", true);
    assertEquals(13, sieve.scanLowerCase(text));
}

public void testInexactAtBeginning() {
    String text = "\u4f60\u4eec\u597d\u5417\uff1f"; //Ni3 men hao3 ma?
    SearchSieve sieve = new SearchSieve("\u4f60\u4eec");
    assertEquals(1, sieve.scanLowerCase(text));
}

public void testExactNoMatch() {
    String text = "Lucy loves the cha cha cha.";
    SearchSieve sieve = new SearchSieve("love", true);
    assertEquals(-1, sieve.scanLowerCase(text));
}

public void testInexactNoMatch() {
    String text = "Potatoes killed the Quayle";
    SearchSieve sieve = new SearchSieve("qualia");
    assertEquals(-1, sieve.scanLowerCase(text));
}

public void testContinued() {
    char[] text = "Flying Fishery".toCharArray();
    SearchSieve sieve = new SearchSieve("flying fishery");
    assertEquals(-1, sieve.scanLowerCase(text, 0, 6));
    assertEquals(text.length - 1,
                sieve.scanLowerCase(text, 6, text.length - 6));
}

public void testOffset() {
    char[] text = "haha".toCharArray();
    SearchSieve sieve = new SearchSieve("ha");
    assertEquals(3, sieve.scanAllLowerCase(text, 1, 3));
}

public void testScanAll1() {
    String text = "sousa sous";
    SearchSieve sieve = new SearchSieve("sousa");
    assertEquals(4, sieve.scanAllLowerCase(text));
    assertTrue(sieve.addChar('a'));
}

public void testScanAll2() {
    SearchSieve sieve = new SearchSieve("banana", true);
    assertEquals(6, sieve.scanAllLowerCase("banana ban ban ban"));
    assertEquals(3, sieve.scanAllLowerCase("ana\u3000"));
}

public void testHasMatched() {
    SearchSieve sieve = new SearchSieve("hasmatched");
    sieve.scanAllLowerCase("boolean hasMatched() ");
    assertTrue(sieve.hasMatched());
}

}